@@ -1,7 +1,7 @@
 /* LICENSE BEGIN
     This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2020 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2020 Simon Hausmann <simon.hausmann@sixtyfps.io>
+    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
+    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
 
     SPDX-License-Identifier: GPL-3.0-only
     This file is also available under commercial licensing terms.
@@ -12,6 +12,7 @@ use core::convert::TryInto;
 use sixtyfps_compilerlib::langtype::Type as LangType;
 use sixtyfps_corelib::graphics::Image;
 use sixtyfps_corelib::{Brush, PathData, SharedString, SharedVector};
+use std::borrow::Cow;
 use std::collections::HashMap;
 use std::iter::FromIterator;
 use std::path::{Path, PathBuf};
@@ -20,6 +21,8 @@ use std::rc::Rc;
 #[doc(inline)]
 pub use sixtyfps_compilerlib::diagnostics::{Diagnostic, DiagnosticLevel};
 
+pub use sixtyfps_corelib::window::api::Window;
+
 /// This enum represents the different public variants of the [`Value`] enum, without
 /// the contained values.
 // NOTE: The docs for ValueType are duplicated in sixtyfps_interpreter.h, for extraction by
@@ -66,6 +69,7 @@ impl From<LangType> for ValueType {
             LangType::Bool => Self::Bool,
             LangType::Struct { .. } => Self::Struct,
             LangType::Void => Self::Void,
+            LangType::Image => Self::Image,
             _ => Self::Other,
         }
     }
@@ -95,7 +99,6 @@ pub enum Value {
     String(SharedString),
     /// Correspond to the `bool` type in .60
     Bool(bool),
-    #[doc(hidden)]
     /// Correspond to the `image` type in .60
     Image(Image),
     /// An Array in the .60 language.
@@ -132,6 +135,7 @@ impl Value {
             Value::Model(_) => ValueType::Model,
             Value::Struct(_) => ValueType::Struct,
             Value::Brush(_) => ValueType::Brush,
+            Value::Image(_) => ValueType::Image,
             _ => ValueType::Other,
         }
     }
@@ -226,9 +230,9 @@ declare_value_conversion!(LayoutCache => [SharedVector<f32>] );
 
 /// Implement From / TryInto for Value that convert a `struct` to/from `Value::Object`
 macro_rules! declare_value_struct_conversion {
-    (struct $name:path { $($field:ident),* $(,)? }) => {
+    (struct $name:path { $($field:ident),* $(, ..$extra:expr)? }) => {
         impl From<$name> for Value {
-            fn from($name { $($field),* }: $name) -> Self {
+            fn from($name { $($field),* , .. }: $name) -> Self {
                 let mut struct_ = Struct::default();
                 $(struct_.set_field(stringify!($field).into(), $field.into());)*
                 Value::Struct(struct_)
@@ -242,6 +246,7 @@ macro_rules! declare_value_struct_conversion {
                         type Ty = $name;
                         Ok(Ty {
                             $($field: x.get_field(stringify!($field)).ok_or(())?.clone().try_into().map_err(|_|())?),*
+                            $(, ..$extra)?
                         })
                     }
                     _ => Err(()),
@@ -256,6 +261,7 @@ declare_value_struct_conversion!(struct sixtyfps_corelib::properties::StateInfo
 declare_value_struct_conversion!(struct sixtyfps_corelib::input::KeyboardModifiers { control, alt, shift, meta });
 declare_value_struct_conversion!(struct sixtyfps_corelib::input::KeyEvent { event_type, text, modifiers });
 declare_value_struct_conversion!(struct sixtyfps_corelib::layout::LayoutInfo { min, max, min_percent, max_percent, preferred, stretch });
+declare_value_struct_conversion!(struct sixtyfps_corelib::graphics::Point { x, y, ..Default::default()});
 
 /// Implement From / TryInto for Value that convert an `enum` to/from `Value::EnumerationValue`
 ///
@@ -265,7 +271,7 @@ macro_rules! declare_value_enum_conversion {
     ($ty:ty, $n:ident) => {
         impl From<$ty> for Value {
             fn from(v: $ty) -> Self {
-                Value::EnumerationValue(stringify!($n).to_owned(), v.to_string())
+                Value::EnumerationValue(stringify!($n).to_owned(), v.to_string().replace('_', "-"))
             }
         }
         impl TryInto<$ty> for Value {
@@ -278,7 +284,9 @@ macro_rules! declare_value_enum_conversion {
                             return Err(());
                         }
 
-                        <$ty>::from_str(value.as_str()).map_err(|_| ())
+                        <$ty>::from_str(value.as_str())
+                            .or_else(|_| <$ty>::from_str(&value.as_str().replace('-', "_")))
+                            .map_err(|_| ())
                     }
                     _ => Err(()),
                 }
@@ -299,6 +307,7 @@ declare_value_enum_conversion!(sixtyfps_corelib::items::TextOverflow, TextOverfl
 declare_value_enum_conversion!(sixtyfps_corelib::items::TextWrap, TextWrap);
 declare_value_enum_conversion!(sixtyfps_corelib::layout::LayoutAlignment, LayoutAlignment);
 declare_value_enum_conversion!(sixtyfps_corelib::items::ImageFit, ImageFit);
+declare_value_enum_conversion!(sixtyfps_corelib::items::ImageRendering, ImageRendering);
 declare_value_enum_conversion!(sixtyfps_corelib::input::KeyEventType, KeyEventType);
 declare_value_enum_conversion!(sixtyfps_corelib::items::EventResult, EventResult);
 declare_value_enum_conversion!(sixtyfps_corelib::items::FillRule, FillRule);
@@ -349,6 +358,15 @@ impl TryInto<sixtyfps_corelib::Color> for Value {
     }
 }
 
+/// Normalize the identifier to use dashes
+pub(crate) fn normalize_identifier(ident: &str) -> Cow<'_, str> {
+    if ident.contains('_') {
+        ident.replace('_', "-").into()
+    } else {
+        ident.into()
+    }
+}
+
 /// This type represents a runtime instance of structure in `.60`.
 ///
 /// This can either be an instance of a name structure introduced
@@ -375,11 +393,15 @@ pub struct Struct(HashMap<String, Value>);
 impl Struct {
     /// Get the value for a given struct field
     pub fn get_field(&self, name: &str) -> Option<&Value> {
-        self.0.get(name)
+        self.0.get(&*normalize_identifier(name))
     }
     /// Set the value of a given struct field
     pub fn set_field(&mut self, name: String, value: Value) {
-        self.0.insert(name, value);
+        if name.contains('_') {
+            self.0.insert(name.replace('_', "-"), value);
+        } else {
+            self.0.insert(name, value);
+        }
     }
 
     /// Iterate over all the fields in this struct
@@ -390,7 +412,11 @@ impl Struct {
 
 impl FromIterator<(String, Value)> for Struct {
     fn from_iter<T: IntoIterator<Item = (String, Value)>>(iter: T) -> Self {
-        Self(iter.into_iter().collect())
+        Self(
+            iter.into_iter()
+                .map(|(s, v)| (if s.contains('_') { s.replace('_', "-") } else { s }, v))
+                .collect(),
+        )
     }
 }
 
@@ -418,9 +444,8 @@ pub struct ComponentCompiler {
     diagnostics: Vec<Diagnostic>,
 }
 
-impl ComponentCompiler {
-    /// Returns a new ComponentCompiler
-    pub fn new() -> Self {
+impl Default for ComponentCompiler {
+    fn default() -> Self {
         Self {
             config: sixtyfps_compilerlib::CompilerConfiguration::new(
                 sixtyfps_compilerlib::generator::OutputFormat::Interpreter,
@@ -428,6 +453,13 @@ impl ComponentCompiler {
             diagnostics: vec![],
         }
     }
+}
+
+impl ComponentCompiler {
+    /// Returns a new ComponentCompiler.
+    pub fn new() -> Self {
+        Self::default()
+    }
 
     /// Sets the include paths used for looking up `.60` imports to the specified vector of paths.
     pub fn set_include_paths(&mut self, include_paths: Vec<std::path::PathBuf>) {
@@ -468,7 +500,7 @@ impl ComponentCompiler {
             Some(Rc::new(move |path| file_loader_fallback(Path::new(path.as_str()))));
     }
 
-    /// Returns the diagnostics that were produced in the last call to [[Self::build_from_path]] or [[Self::build_from_source]].
+    /// Returns the diagnostics that were produced in the last call to [`Self::build_from_path`] or [`Self::build_from_source`].
     pub fn diagnostics(&self) -> &Vec<Diagnostic> {
         &self.diagnostics
     }
@@ -478,7 +510,7 @@ impl ComponentCompiler {
     /// Returns the compiled `ComponentDefinition` if there were no errors.
     ///
     /// Any diagnostics produced during the compilation, such as warnings or errors, are collected
-    /// in this ComponentCompiler and can be retrieved after the call using the [[Self::diagnostics()]]
+    /// in this ComponentCompiler and can be retrieved after the call using the [`Self::diagnostics()`]
     /// function. The [`print_diagnostics`] function can be used to display the diagnostics
     /// to the users.
     ///
@@ -514,7 +546,7 @@ impl ComponentCompiler {
     /// paths while importing.
     ///
     /// Any diagnostics produced during the compilation, such as warnings or errors, are collected
-    /// in this ComponentCompiler and can be retrieved after the call using the [[Self::diagnostics()]]
+    /// in this ComponentCompiler and can be retrieved after the call using the [`Self::diagnostics()`]
     /// function. The [`print_diagnostics`] function can be used to display the diagnostics
     /// to the users.
     ///
@@ -569,15 +601,17 @@ impl ComponentDefinition {
     }
 
     /// Instantiate the component using an existing window.
-    /// This method is internal because the Window is not a public type
+    /// This method is internal because the WindowRc is not a public type
     #[doc(hidden)]
-    pub fn create_with_existing_window(
-        &self,
-        window: sixtyfps_corelib::window::ComponentWindow,
-    ) -> ComponentInstance {
+    pub fn create_with_existing_window(&self, window: &Window) -> ComponentInstance {
+        use sixtyfps_corelib::window::WindowHandleAccess;
         generativity::make_guard!(guard);
         ComponentInstance {
-            inner: self.inner.unerase(guard).clone().create_with_existing_window(window),
+            inner: self
+                .inner
+                .unerase(guard)
+                .clone()
+                .create_with_existing_window(window.window_handle()),
         }
     }
 
@@ -595,7 +629,7 @@ impl ComponentDefinition {
     }
 
     /// List of publicly declared properties.
-    pub fn properties<'a>(&'a self) -> impl Iterator<Item = (String, ValueType)> + 'a {
+    pub fn properties(&self) -> impl Iterator<Item = (String, ValueType)> + '_ {
         // We create here a 'static guard, because unfortunately the returned type would be restricted to the guard lifetime
         // which is not required, but this is safe because there is only one instance of the unerased type
         let guard = unsafe { generativity::Guard::new(generativity::Id::new()) };
@@ -609,7 +643,7 @@ impl ComponentDefinition {
     }
 
     /// Returns the names of all publicly declared callbacks.
-    pub fn callbacks<'a>(&'a self) -> impl Iterator<Item = String> + 'a {
+    pub fn callbacks(&self) -> impl Iterator<Item = String> + '_ {
         // We create here a 'static guard, because unfortunately the returned type would be restricted to the guard lifetime
         // which is not required, but this is safe because there is only one instance of the unerased type
         let guard = unsafe { generativity::Guard::new(generativity::Id::new()) };
@@ -678,7 +712,7 @@ impl ComponentInstance {
     ///         property <int> my_property: 42;
     ///     }
     /// "#;
-    /// let mut compiler = ComponentCompiler::new();
+    /// let mut compiler = ComponentCompiler::default();
     /// let definition = spin_on::spin_on(
     ///     compiler.build_from_source(code.into(), Default::default()));
     /// assert!(compiler.diagnostics().is_empty(), "{:?}", compiler.diagnostics());
@@ -689,7 +723,7 @@ impl ComponentInstance {
         generativity::make_guard!(guard);
         let comp = self.inner.unerase(guard);
         comp.description()
-            .get_property(comp.borrow(), name)
+            .get_property(comp.borrow(), &normalize_identifier(name))
             .map_err(|()| GetPropertyError::NoSuchProperty)
     }
 
@@ -697,9 +731,7 @@ impl ComponentInstance {
     pub fn set_property(&self, name: &str, value: Value) -> Result<(), SetPropertyError> {
         generativity::make_guard!(guard);
         let comp = self.inner.unerase(guard);
-        comp.description()
-            .set_property(comp.borrow(), name, value)
-            .map_err(|()| todo!("set_property don't return the right error type"))
+        comp.description().set_property(comp.borrow(), &normalize_identifier(name), value)
     }
 
     /// Set a handler for the callback with the given name. A callback with that
@@ -721,7 +753,7 @@ impl ComponentInstance {
     ///     }
     /// "#;
     /// let definition = spin_on::spin_on(
-    ///     ComponentCompiler::new().build_from_source(code.into(), Default::default()));
+    ///     ComponentCompiler::default().build_from_source(code.into(), Default::default()));
     /// let instance = definition.unwrap().create();
     ///
     /// let instance_weak = instance.as_weak();
@@ -743,7 +775,7 @@ impl ComponentInstance {
         generativity::make_guard!(guard);
         let comp = self.inner.unerase(guard);
         comp.description()
-            .set_callback_handler(comp.borrow(), name, Box::new(callback))
+            .set_callback_handler(comp.borrow(), &normalize_identifier(name), Box::new(callback))
             .map_err(|()| SetCallbackError::NoSuchCallback)
     }
 
@@ -754,7 +786,133 @@ impl ComponentInstance {
     pub fn invoke_callback(&self, name: &str, args: &[Value]) -> Result<Value, CallCallbackError> {
         generativity::make_guard!(guard);
         let comp = self.inner.unerase(guard);
-        Ok(comp.description().invoke_callback(comp.borrow(), name, &args).map_err(|()| todo!())?)
+        comp.description()
+            .invoke_callback(comp.borrow(), &normalize_identifier(name), args)
+            .map_err(|()| CallCallbackError::NoSuchCallback)
+    }
+
+    /// Return the value for a property within an exported global singleton used by this component.
+    ///
+    /// The `global` parameter is the exported name of the global singleton. The `property` argument
+    /// is the name of the property
+    ///
+    /// ## Examples
+    ///
+    /// ```
+    /// use sixtyfps_interpreter::{ComponentDefinition, ComponentCompiler, Value, SharedString};
+    /// let code = r#"
+    ///     global Glob := {
+    ///         property <int> my_property: 42;
+    ///     }
+    ///     export { Glob as TheGlobal }
+    ///     MyWin := Window {
+    ///     }
+    /// "#;
+    /// let mut compiler = ComponentCompiler::default();
+    /// let definition = spin_on::spin_on(
+    ///     compiler.build_from_source(code.into(), Default::default()));
+    /// assert!(compiler.diagnostics().is_empty(), "{:?}", compiler.diagnostics());
+    /// let instance = definition.unwrap().create();
+    /// assert_eq!(instance.get_global_property("TheGlobal", "my_property").unwrap(), Value::from(42));
+    /// ```
+    pub fn get_global_property(
+        &self,
+        global: &str,
+        property: &str,
+    ) -> Result<Value, GetPropertyError> {
+        generativity::make_guard!(guard);
+        let comp = self.inner.unerase(guard);
+        comp.description()
+            .get_global(comp.borrow(), &normalize_identifier(global))
+            .map_err(|()| GetPropertyError::NoSuchProperty)? // FIXME: should there be a NoSuchGlobal error?
+            .as_ref()
+            .get_property(&normalize_identifier(property))
+            .map_err(|()| GetPropertyError::NoSuchProperty)
+    }
+
+    /// Set the value for a property within an exported global singleton used by this component.
+    pub fn set_global_property(
+        &self,
+        global: &str,
+        property: &str,
+        value: Value,
+    ) -> Result<(), SetPropertyError> {
+        generativity::make_guard!(guard);
+        let comp = self.inner.unerase(guard);
+        comp.description()
+            .get_global(comp.borrow(), &normalize_identifier(global))
+            .map_err(|()| SetPropertyError::NoSuchProperty)? // FIXME: should there be a NoSuchGlobal error?
+            .as_ref()
+            .set_property(&normalize_identifier(property), value)
+    }
+
+    /// Set a handler for the callback in the exported global singleton. A callback with that
+    /// name must be defined in the specified global and the global must be exported from the
+    /// main document otherwise an error will be returned.
+    ///
+    /// ## Examples
+    ///
+    /// ```
+    /// use sixtyfps_interpreter::{ComponentDefinition, ComponentCompiler, Value, SharedString};
+    /// use core::convert::TryInto;
+    /// let code = r#"
+    ///     export global Logic := {
+    ///         callback to_uppercase(string) -> string;
+    ///     }
+    ///     MyWin := Window {
+    ///         property <string> hello: Logic.to_uppercase("world");
+    ///     }
+    /// "#;
+    /// let definition = spin_on::spin_on(
+    ///     ComponentCompiler::default().build_from_source(code.into(), Default::default()));
+    /// let instance = definition.unwrap().create();
+    /// instance.set_global_callback("Logic", "to_uppercase", |args: &[Value]| -> Value {
+    ///     let arg: SharedString = args[0].clone().try_into().unwrap();
+    ///     Value::from(SharedString::from(arg.to_uppercase()))
+    /// }).unwrap();
+    ///
+    /// let res = instance.get_property("hello").unwrap();
+    /// assert_eq!(res, Value::from(SharedString::from("WORLD")));
+    ///
+    /// let abc = instance.invoke_global_callback("Logic", "to_uppercase", &[
+    ///     SharedString::from("abc").into()
+    /// ]).unwrap();
+    /// assert_eq!(abc, Value::from(SharedString::from("ABC")));
+    /// ```
+    pub fn set_global_callback(
+        &self,
+        global: &str,
+        name: &str,
+        callback: impl Fn(&[Value]) -> Value + 'static,
+    ) -> Result<(), SetCallbackError> {
+        generativity::make_guard!(guard);
+        let comp = self.inner.unerase(guard);
+        comp.description()
+            .get_global(comp.borrow(), &normalize_identifier(global))
+            .map_err(|()| SetCallbackError::NoSuchCallback)? // FIXME: should there be a NoSuchGlobal error?
+            .as_ref()
+            .set_callback_handler(&normalize_identifier(name), Box::new(callback))
+            .map_err(|()| SetCallbackError::NoSuchCallback)
+    }
+
+    /// Call the given callback within a global singleton with the arguments
+    ///
+    /// ## Examples
+    /// See the documentation of [`Self::set_global_callback`] for an example
+    pub fn invoke_global_callback(
+        &self,
+        global: &str,
+        callback_name: &str,
+        args: &[Value],
+    ) -> Result<Value, CallCallbackError> {
+        generativity::make_guard!(guard);
+        let comp = self.inner.unerase(guard);
+        comp.description()
+            .get_global(comp.borrow(), &normalize_identifier(global))
+            .map_err(|()| CallCallbackError::NoSuchCallback)? // FIXME: should there be a NoSuchGlobal error?
+            .as_ref()
+            .invoke_callback(&normalize_identifier(callback_name), args)
+            .map_err(|()| CallCallbackError::NoSuchCallback)
     }
 
     /// Marks the window of this component to be shown on the screen. This registers
@@ -764,7 +922,7 @@ impl ComponentInstance {
     pub fn show(&self) {
         generativity::make_guard!(guard);
         let comp = self.inner.unerase(guard);
-        comp.window().show();
+        comp.borrow_instance().window().show();
     }
 
     /// Marks the window of this component to be hidden on the screen. This de-registers
@@ -772,8 +930,9 @@ impl ComponentInstance {
     pub fn hide(&self) {
         generativity::make_guard!(guard);
         let comp = self.inner.unerase(guard);
-        comp.window().hide();
+        comp.borrow_instance().window().hide();
     }
+
     /// This is a convenience function that first calls [`Self::show`], followed by [`crate::run_event_loop()`]
     /// and [`Self::hide`].
     pub fn run(&self) {
@@ -800,13 +959,11 @@ impl ComponentInstance {
         WeakComponentInstance { inner: vtable::VRc::downgrade(&self.inner) }
     }
 
-    /// Return the window.
-    /// This method is internal because the Window is not a public type
-    #[doc(hidden)]
-    pub fn window(&self) -> sixtyfps_corelib::window::ComponentWindow {
-        generativity::make_guard!(guard);
-        let comp = self.inner.unerase(guard);
-        comp.window()
+    /// Returns the Window associated with this component. The window API can be used
+    /// to control different aspects of the integration into the windowing system,
+    /// such as the position on the screen.
+    pub fn window(&self) -> &Window {
+        self.inner.window()
     }
 }
 
@@ -879,13 +1036,15 @@ pub fn run_event_loop() {
 
 /// This module contains a few function use by tests
 pub mod testing {
+    use sixtyfps_corelib::window::WindowHandleAccess;
+
     /// Wrapper around [`sixtyfps_corelib::tests::sixtyfps_send_mouse_click`]
     pub fn send_mouse_click(comp: &super::ComponentInstance, x: f32, y: f32) {
         sixtyfps_corelib::tests::sixtyfps_send_mouse_click(
             &vtable::VRc::into_dyn(comp.inner.clone()),
             x,
             y,
-            &comp.inner.window(),
+            comp.window().window_handle(),
         );
     }
     /// Wrapper around [`sixtyfps_corelib::tests::send_keyboard_string_sequence`]
@@ -896,19 +1055,20 @@ pub mod testing {
         sixtyfps_corelib::tests::send_keyboard_string_sequence(
             &string,
             Default::default(),
-            &comp.inner.window(),
+            comp.window().window_handle(),
         );
     }
 }
 
 #[test]
 fn component_definition_properties() {
-    let mut compiler = ComponentCompiler::new();
+    let mut compiler = ComponentCompiler::default();
     let comp_def = spin_on::spin_on(
         compiler.build_from_source(
             r#"
     export Dummy := Rectangle {
         property <string> test;
+        property <int> underscores-and-dashes_preserved: 44;
         callback hello;
     }"#
             .into(),
@@ -919,14 +1079,36 @@ fn component_definition_properties() {
 
     let props = comp_def.properties().collect::<Vec<(_, _)>>();
 
-    assert_eq!(props.len(), 1);
+    assert_eq!(props.len(), 2);
     assert_eq!(props[0].0, "test");
     assert_eq!(props[0].1, ValueType::String);
+    assert_eq!(props[1].0, "underscores-and-dashes_preserved");
+    assert_eq!(props[1].1, ValueType::Number);
+
+    let instance = comp_def.create();
+    assert_eq!(instance.get_property("underscores_and-dashes-preserved"), Ok(Value::Number(44.)));
+    assert_eq!(
+        instance.get_property("underscoresanddashespreserved"),
+        Err(GetPropertyError::NoSuchProperty)
+    );
+    assert_eq!(
+        instance.set_property("underscores-and_dashes-preserved", Value::Number(88.)),
+        Ok(())
+    );
+    assert_eq!(
+        instance.set_property("underscoresanddashespreserved", Value::Number(99.)),
+        Err(SetPropertyError::NoSuchProperty)
+    );
+    assert_eq!(
+        instance.set_property("underscores-and_dashes-preserved", Value::String("99".into())),
+        Err(SetPropertyError::WrongType)
+    );
+    assert_eq!(instance.get_property("underscores-and-dashes-preserved"), Ok(Value::Number(88.)));
 }
 
 #[test]
 fn component_definition_properties2() {
-    let mut compiler = ComponentCompiler::new();
+    let mut compiler = ComponentCompiler::default();
     let comp_def = spin_on::spin_on(
         compiler.build_from_source(
             r#"
@@ -944,7 +1126,7 @@ fn component_definition_properties2() {
     let props = comp_def.properties().collect::<Vec<(_, _)>>();
 
     assert_eq!(props.len(), 1);
-    assert_eq!(props[0].0, "sub_text");
+    assert_eq!(props[0].0, "sub-text");
     assert_eq!(props[0].1, ValueType::String);
 
     let callbacks = comp_def.callbacks().collect::<Vec<_>>();
@@ -952,6 +1134,90 @@ fn component_definition_properties2() {
     assert_eq!(callbacks[0], "hello");
 }
 
+#[test]
+fn globals() {
+    let mut compiler = ComponentCompiler::default();
+    let definition = spin_on::spin_on(
+        compiler.build_from_source(
+            r#"
+    export global My-Super_Global := {
+        property <int> the-property : 21;
+    }
+    export { My-Super_Global as AliasedGlobal }
+    export Dummy := Rectangle {
+    }"#
+            .into(),
+            "".into(),
+        ),
+    );
+    let instance = definition.unwrap().create();
+    assert_eq!(
+        instance.set_global_property("My_Super-Global", "the_property", Value::Number(44.)),
+        Ok(())
+    );
+    assert_eq!(
+        instance.set_global_property("AliasedGlobal", "the_property", Value::Number(44.)),
+        Ok(())
+    );
+    assert_eq!(
+        instance.set_global_property("DontExist", "the-property", Value::Number(88.)),
+        Err(SetPropertyError::NoSuchProperty)
+    );
+
+    assert_eq!(
+        instance.set_global_property("My_Super-Global", "theproperty", Value::Number(88.)),
+        Err(SetPropertyError::NoSuchProperty)
+    );
+    assert_eq!(
+        instance.set_global_property("AliasedGlobal", "theproperty", Value::Number(88.)),
+        Err(SetPropertyError::NoSuchProperty)
+    );
+    assert_eq!(
+        instance.set_global_property("My_Super-Global", "the_property", Value::String("88".into())),
+        Err(SetPropertyError::WrongType)
+    );
+    assert_eq!(
+        instance.get_global_property("My-Super_Global", "yoyo"),
+        Err(GetPropertyError::NoSuchProperty)
+    );
+    assert_eq!(
+        instance.get_global_property("My-Super_Global", "the-property"),
+        Ok(Value::Number(44.))
+    );
+
+    assert_eq!(
+        instance.set_property("the-property", Value::Void),
+        Err(SetPropertyError::NoSuchProperty)
+    );
+    assert_eq!(instance.get_property("the-property"), Err(GetPropertyError::NoSuchProperty));
+
+    assert_eq!(
+        instance.set_global_callback("DontExist", "the-property", |_| panic!()),
+        Err(SetCallbackError::NoSuchCallback)
+    );
+    assert_eq!(
+        instance.set_global_callback("My_Super_Global", "the-property", |_| panic!()),
+        Err(SetCallbackError::NoSuchCallback)
+    );
+    assert_eq!(
+        instance.set_global_callback("My_Super_Global", "yoyo", |_| panic!()),
+        Err(SetCallbackError::NoSuchCallback)
+    );
+
+    assert_eq!(
+        instance.invoke_global_callback("DontExist", "the-property", &[]),
+        Err(CallCallbackError::NoSuchCallback)
+    );
+    assert_eq!(
+        instance.invoke_global_callback("My_Super_Global", "the-property", &[]),
+        Err(CallCallbackError::NoSuchCallback)
+    );
+    assert_eq!(
+        instance.invoke_global_callback("My_Super_Global", "yoyo", &[]),
+        Err(CallCallbackError::NoSuchCallback)
+    );
+}
+
 #[cfg(feature = "ffi")]
 #[allow(missing_docs)]
 #[path = "ffi.rs"]
