@@ -1,7 +1,7 @@
 /* LICENSE BEGIN
     This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2020 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2020 Simon Hausmann <simon.hausmann@sixtyfps.io>
+    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
+    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
 
     SPDX-License-Identifier: GPL-3.0-only
     This file is also available under commercial licensing terms.
@@ -14,7 +14,7 @@ use core::ptr::NonNull;
 use dynamic_type::{Instance, InstanceBox};
 use sixtyfps_compilerlib::expression_tree::{Expression, NamedReference};
 use sixtyfps_compilerlib::langtype::Type;
-use sixtyfps_compilerlib::object_tree::{Element, ElementRc};
+use sixtyfps_compilerlib::object_tree::ElementRc;
 use sixtyfps_compilerlib::*;
 use sixtyfps_compilerlib::{diagnostics::BuildDiagnostics, object_tree::PropertyDeclaration};
 use sixtyfps_corelib::component::{Component, ComponentRef, ComponentRefPin, ComponentVTable};
@@ -30,7 +30,7 @@ use sixtyfps_corelib::model::Repeater;
 use sixtyfps_corelib::properties::InterpolatedPropertyValue;
 use sixtyfps_corelib::rtti::{self, AnimatedBindingKind, FieldOffset, PropertyInfo};
 use sixtyfps_corelib::slice::Slice;
-use sixtyfps_corelib::window::ComponentWindow;
+use sixtyfps_corelib::window::{api::Window, WindowHandleAccess, WindowRc};
 use sixtyfps_corelib::{Brush, Color, Property, SharedString, SharedVector};
 use std::collections::BTreeMap;
 use std::collections::HashMap;
@@ -43,47 +43,43 @@ pub struct ComponentBox<'id> {
 
 impl<'id> ComponentBox<'id> {
     /// Borrow this component as a `Pin<ComponentRef>`
-    pub fn borrow<'a>(&'a self) -> ComponentRefPin<'a> {
+    pub fn borrow(&self) -> ComponentRefPin {
         self.borrow_instance().borrow()
     }
 
     /// Safety: the lifetime is not unique
     pub fn description(&self) -> Rc<ComponentDescription<'id>> {
-        return self.component_type.clone();
+        self.component_type.clone()
     }
 
     pub fn borrow_instance<'a>(&'a self) -> InstanceRef<'a, 'id> {
         InstanceRef { instance: self.instance.as_pin_ref(), component_type: &self.component_type }
     }
 
-    pub fn window(&self) -> ComponentWindow {
-        (*self
-            .component_type
+    pub fn window(&self) -> &Window {
+        self.component_type
             .window_offset
-            .apply_pin(self.instance.as_pin_ref())
-            .as_pin_ref()
-            .unwrap())
-        .clone()
+            .apply(self.instance.as_pin_ref().get_ref())
+            .as_ref()
+            .as_ref()
+            .unwrap()
     }
 }
 
 impl<'id> Drop for ComponentBox<'id> {
     fn drop(&mut self) {
         let instance_ref = self.borrow_instance();
-        match eval::window_ref(instance_ref) {
-            Some(window) => {
-                let items = self
-                    .component_type
-                    .items
-                    .values()
-                    .map(|item_within_component| unsafe {
-                        item_within_component.item_from_component(instance_ref.as_ptr())
-                    })
-                    .collect::<Vec<_>>();
+        if let Some(window) = eval::window_ref(instance_ref) {
+            let items = self
+                .component_type
+                .items
+                .values()
+                .map(|item_within_component| unsafe {
+                    item_within_component.item_from_component(instance_ref.as_ptr())
+                })
+                .collect::<Vec<_>>();
 
-                window.free_graphics_resources(&Slice::from_slice(items.as_slice()));
-            }
-            None => {}
+            window.free_graphics_resources(&Slice::from_slice(items.as_slice()));
         }
     }
 }
@@ -181,7 +177,7 @@ impl Component for ErasedComponentBox {
     ) -> sixtyfps_corelib::layout::LayoutInfo {
         self.borrow().as_ref().layout_info(orientation)
     }
-    fn get_item_ref<'a>(self: Pin<&'a Self>, index: usize) -> Pin<ItemRef<'a>> {
+    fn get_item_ref(self: Pin<&Self>, index: usize) -> Pin<ItemRef> {
         // We're having difficulties transferring the lifetime to a pinned reference
         // to the other ComponentVTable with the same life time. So skip the vtable
         // indirection and call our implementation directly.
@@ -294,8 +290,7 @@ pub struct ComponentDescription<'id> {
     pub(crate) parent_component_offset:
         Option<FieldOffset<Instance<'id>, Option<ComponentRefPin<'id>>>>,
     /// Offset to the window reference
-    pub(crate) window_offset:
-        FieldOffset<Instance<'id>, Option<sixtyfps_corelib::window::ComponentWindow>>,
+    pub(crate) window_offset: FieldOffset<Instance<'id>, Option<Window>>,
     /// Offset of a ComponentExtraData
     pub(crate) extra_data_offset: FieldOffset<Instance<'id>, ComponentExtraData>,
     /// Keep the Rc alive
@@ -314,13 +309,24 @@ impl<'id> ComponentDescription<'id> {
     }
 
     /// List of publicly declared properties or callbacks
+    ///
+    /// We try to preserve the dashes and underscore as written in the property declaration
     pub fn properties(
         &self,
     ) -> impl Iterator<Item = (String, sixtyfps_compilerlib::langtype::Type)> + '_ {
-        self.public_properties
-            .iter()
-            .filter(|(_, v)| v.expose_in_public_api)
-            .map(|(s, v)| (s.clone(), v.property_type.clone()))
+        self.public_properties.iter().filter(|(_, v)| v.expose_in_public_api).map(|(s, v)| {
+            let name = v
+                .node
+                .as_ref()
+                .and_then(|n| {
+                    n.as_ref()
+                        .either(|n| n.DeclaredIdentifier(), |n| n.DeclaredIdentifier())
+                        .child_token(parser::SyntaxKind::Identifier)
+                })
+                .map(|n| n.to_string())
+                .unwrap_or_else(|| s.clone());
+            (name, v.property_type.clone())
+        })
     }
 
     /// Instantiate a runtime component from this ComponentDescription
@@ -336,18 +342,19 @@ impl<'id> ComponentDescription<'id> {
             sixtyfps_rendering_backend_default::backend();
             sixtyfps_rendering_backend_gl::create_gl_window_with_canvas_id(canvas_id)
         };
-        self.create_with_existing_window(window)
+        self.create_with_existing_window(&window)
     }
 
     #[doc(hidden)]
     pub fn create_with_existing_window(
         self: Rc<Self>,
-        window: sixtyfps_corelib::window::ComponentWindow,
+        window: &sixtyfps_corelib::window::WindowRc,
     ) -> vtable::VRc<ComponentVTable, ErasedComponentBox> {
         let component_ref = instantiate(self, None, Some(window));
         component_ref
             .as_pin_ref()
             .window()
+            .window_handle()
             .set_component(&vtable::VRc::into_dyn(component_ref.clone()));
         component_ref.run_setup_code();
         component_ref
@@ -355,16 +362,17 @@ impl<'id> ComponentDescription<'id> {
 
     /// Set a value to property.
     ///
-    /// Returns an error if the component is not an instance corresponding to this ComponentDescription,
-    /// or if the property with this name does not exist in this component
+    /// Return an error if the property with this name does not exist in this component,
+    /// or if the value is the wrong type.
+    /// Panics if the component is not an instance corresponding to this ComponentDescription,
     pub fn set_property(
         &self,
         component: ComponentRefPin,
         name: &str,
         value: Value,
-    ) -> Result<(), ()> {
+    ) -> Result<(), crate::api::SetPropertyError> {
         if !core::ptr::eq((&self.ct) as *const _, component.get_vtable() as *const _) {
-            return Err(());
+            panic!("mismatch instance and vtable");
         }
         generativity::make_guard!(guard);
         let c = unsafe { InstanceRef::from_pin_ref(component, guard) };
@@ -523,6 +531,22 @@ impl<'id> ComponentDescription<'id> {
             .ok_or(())
         }
     }
+
+    // Return the global with the given name
+    pub fn get_global(
+        &self,
+        component: ComponentRefPin,
+        global_name: &str,
+    ) -> Result<Pin<Rc<dyn crate::global_component::GlobalComponent>>, ()> {
+        if !core::ptr::eq((&self.ct) as *const _, component.get_vtable() as *const _) {
+            return Err(());
+        }
+        generativity::make_guard!(guard);
+        // Safety: we just verified that the component has the right vtable
+        let c = unsafe { InstanceRef::from_pin_ref(component, guard) };
+        let extra_data = c.component_type.extra_data_offset.apply(c.instance.get_ref());
+        extra_data.globals.get(global_name).map(|g| g.clone()).ok_or(())
+    }
 }
 
 extern "C" fn visit_children_item(
@@ -537,7 +561,7 @@ extern "C" fn visit_children_item(
     sixtyfps_corelib::item_tree::visit_item_tree(
         instance_ref.instance,
         &vtable::VRc::into_dyn(comp_rc),
-        instance_ref.component_type.item_tree.as_slice().into(),
+        instance_ref.component_type.item_tree.as_slice(),
         index,
         order,
         v,
@@ -564,12 +588,11 @@ fn ensure_repeater_updated<'id>(
             .window_offset
             .apply(instance_ref.as_ref())
             .as_ref()
-            .unwrap()
-            .clone();
+            .unwrap();
         let instance = instantiate(
             rep_in_comp.component_to_repeat.clone(),
             Some(instance_ref.borrow()),
-            Some(window),
+            Some(window.window_handle()),
         );
         instance.run_setup_code();
         instance
@@ -632,19 +655,19 @@ fn rtti_for<T: 'static + Default + rtti::BuiltinItem + vtable::HasStaticVTable<I
 /// Create a ComponentDescription from a source.
 /// The path corresponding to the source need to be passed as well (path is used for diagnostics
 /// and loading relative assets)
-pub async fn load<'id>(
+pub async fn load(
     source: String,
     path: std::path::PathBuf,
     mut compiler_config: CompilerConfiguration,
-    guard: generativity::Guard<'id>,
-) -> (Result<Rc<ComponentDescription<'id>>, ()>, sixtyfps_compilerlib::diagnostics::BuildDiagnostics)
+    guard: generativity::Guard<'_>,
+) -> (Result<Rc<ComponentDescription<'_>>, ()>, sixtyfps_compilerlib::diagnostics::BuildDiagnostics)
 {
     if compiler_config.style.is_none() && std::env::var("SIXTYFPS_STYLE").is_err() {
         // Defaults to native if it exists:
         compiler_config.style = Some(if sixtyfps_rendering_backend_default::HAS_NATIVE_STYLE {
             "native".to_owned()
         } else {
-            "ugly".to_owned()
+            "fluent".to_owned()
         });
     }
 
@@ -653,10 +676,14 @@ pub async fn load<'id>(
     if diag.has_error() {
         return (Err(()), diag);
     }
-    let (doc, diag) = compile_syntax_node(syntax_node, diag, compiler_config).await;
+    let (doc, mut diag) = compile_syntax_node(syntax_node, diag, compiler_config).await;
     if diag.has_error() {
         return (Err(()), diag);
     }
+    if matches!(doc.root_component.root_element.borrow().base_type, Type::Invalid | Type::Void) {
+        diag.push_error_with_span("No component found".into(), Default::default());
+        return (Err(()), diag);
+    }
     (Ok(generate_component(&doc.root_component, guard)), diag)
 }
 
@@ -664,6 +691,7 @@ pub(crate) fn generate_component<'id>(
     component: &Rc<object_tree::Component>,
     guard: generativity::Guard<'id>,
 ) -> Rc<ComponentDescription<'id>> {
+    //dbg!(&*component.root_element.borrow());
     let mut rtti = HashMap::new();
     {
         use sixtyfps_corelib::items::*;
@@ -678,7 +706,7 @@ pub(crate) fn generate_component<'id>(
                 rtti_for::<FocusScope>(),
                 rtti_for::<Path>(),
                 rtti_for::<Flickable>(),
-                rtti_for::<Window>(),
+                rtti_for::<WindowItem>(),
                 rtti_for::<TextInput>(),
                 rtti_for::<Clip>(),
                 rtti_for::<BoxShadow>(),
@@ -870,14 +898,19 @@ pub(crate) fn generate_component<'id>(
         None
     };
 
-    let window_offset = builder.add_field_type::<Option<ComponentWindow>>();
+    let window_offset = builder.add_field_type::<Option<Window>>();
 
     let extra_data_offset = builder.add_field_type::<ComponentExtraData>();
 
     let public_properties = component.root_element.borrow().property_declarations.clone();
 
-    let compiled_globals =
-        component.used_global.borrow().iter().map(crate::global_component::generate).collect();
+    let compiled_globals = component
+        .used_types
+        .borrow()
+        .globals
+        .iter()
+        .map(crate::global_component::generate)
+        .collect();
 
     let t = ComponentVTable {
         visit_children_item,
@@ -909,10 +942,9 @@ pub(crate) fn generate_component<'id>(
 
 pub fn animation_for_property(
     component: InstanceRef,
-    element: &Element,
-    property_name: &str,
+    animation: &Option<sixtyfps_compilerlib::object_tree::PropertyAnimation>,
 ) -> AnimatedBindingKind {
-    match element.property_animations.get(property_name) {
+    match animation {
         Some(sixtyfps_compilerlib::object_tree::PropertyAnimation::Static(anim_elem)) => {
             AnimatedBindingKind::Animation(eval::new_struct_with_bindings(
                 &anim_elem.borrow().bindings,
@@ -957,7 +989,7 @@ pub fn animation_for_property(
                             );
                         }
                     }
-                    return Default::default();
+                    Default::default()
                 },
             ))
         }
@@ -965,10 +997,10 @@ pub fn animation_for_property(
     }
 }
 
-pub fn instantiate<'id>(
-    component_type: Rc<ComponentDescription<'id>>,
+pub fn instantiate(
+    component_type: Rc<ComponentDescription>,
     parent_ctx: Option<ComponentRefPin>,
-    window: Option<sixtyfps_corelib::window::ComponentWindow>,
+    window: Option<&sixtyfps_corelib::window::WindowRc>,
 ) -> vtable::VRc<ComponentVTable, ErasedComponentBox> {
     let mut instance = component_type.dynamic_type.clone().create_instance();
 
@@ -980,10 +1012,15 @@ pub fn instantiate<'id>(
         extra_data.globals = component_type
             .compiled_globals
             .iter()
-            .map(|g| crate::global_component::instantiate(g))
+            .map(|g| {
+                let (_, instance) = crate::global_component::instantiate(g);
+                g.names().iter().map(|name| (name.clone(), instance.clone())).collect::<Vec<_>>()
+            })
+            .flatten()
             .collect();
     }
-    *component_type.window_offset.apply_mut(instance.as_mut()) = window;
+    *component_type.window_offset.apply_mut(instance.as_mut()) =
+        window.map(|window| window.clone().into());
 
     let component_box = ComponentBox { instance, component_type: component_type.clone() };
     let instance_ref = component_box.borrow_instance();
@@ -991,8 +1028,8 @@ pub fn instantiate<'id>(
     if !component_type.original.is_global() {
         sixtyfps_corelib::component::init_component_items(
             instance_ref.instance,
-            instance_ref.component_type.item_tree.as_slice().into(),
-            &eval::window_ref(instance_ref).unwrap(),
+            instance_ref.component_type.item_tree.as_slice(),
+            eval::window_ref(instance_ref).unwrap(),
         );
     }
 
@@ -1002,7 +1039,7 @@ pub fn instantiate<'id>(
             continue;
         }
         if let Some(b) = component_type.original.root_element.borrow().bindings.get(prop_name) {
-            if !matches!(b.expression, Expression::TwoWayBinding(..)) {
+            if b.two_way_bindings.is_empty() {
                 continue;
             }
         }
@@ -1040,7 +1077,7 @@ pub fn instantiate<'id>(
                         generativity::make_guard!(guard);
                         let mut local_context = eval::EvalLocalContext::from_function_arguments(
                             InstanceRef::from_pin_ref(c, guard),
-                            args.iter().cloned().collect(),
+                            args.to_vec(),
                         );
                         eval::eval_expression(&expr, &mut local_context)
                     })
@@ -1055,7 +1092,7 @@ pub fn instantiate<'id>(
                                 let mut local_context =
                                     eval::EvalLocalContext::from_function_arguments(
                                         InstanceRef::from_pin_ref(c, guard),
-                                        args.iter().cloned().collect(),
+                                        args.to_vec(),
                                     );
                                 eval::eval_expression(&expr, &mut local_context)
                             }),
@@ -1072,10 +1109,7 @@ pub fn instantiate<'id>(
                     component_box.instance.as_ptr().cast(),
                 ));
 
-                let is_state_info = match property_type {
-                    Type::Struct { name: Some(name), .. } if name.ends_with("::StateInfo") => true,
-                    _ => false,
-                };
+                let is_state_info = matches!(property_type, Type::Struct { name: Some(name), .. } if name.ends_with("::StateInfo"));
                 if is_state_info {
                     let prop = Pin::new_unchecked(
                         &*(instance_ref.as_ptr().add(*offset)
@@ -1096,28 +1130,22 @@ pub fn instantiate<'id>(
                     return;
                 }
 
-                let maybe_animation = animation_for_property(
-                    instance_ref,
-                    &component_type.original.root_element.borrow(),
-                    prop_name,
-                );
+                let maybe_animation = animation_for_property(instance_ref, &binding.animation);
                 let item = Pin::new_unchecked(&*instance_ref.as_ptr().add(*offset));
 
-                let mut e = Some(&binding.expression);
-                while let Some(Expression::TwoWayBinding(nr, next)) = &e {
+                for nr in &binding.two_way_bindings {
                     // Safety: The compiler must have ensured that the properties exist and are of the same type
-                    prop_info.link_two_ways(item, get_property_ptr(&nr, instance_ref));
-                    e = next.as_deref();
+                    prop_info.link_two_ways(item, get_property_ptr(nr, instance_ref));
                 }
-                if let Some(e) = e {
-                    if is_const || e.is_constant() {
+                if !matches!(binding.expression, Expression::Invalid) {
+                    if is_const {
                         let v = eval::eval_expression(
-                            e,
+                            &binding.expression,
                             &mut eval::EvalLocalContext::from_component_instance(instance_ref),
                         );
                         prop_info.set(item, v, None).unwrap();
                     } else {
-                        let e = e.clone();
+                        let e = binding.expression.clone();
                         prop_info
                             .set_binding(
                                 item,
@@ -1139,27 +1167,27 @@ pub fn instantiate<'id>(
                 let item_within_component = &component_type.items[&elem.id];
                 let item = item_within_component.item_from_component(instance_ref.as_ptr());
                 if let Some(prop_rtti) = item_within_component.rtti.properties.get(prop_name) {
-                    let maybe_animation = animation_for_property(instance_ref, &elem, prop_name);
-                    let mut e = Some(&binding.expression);
-                    while let Some(Expression::TwoWayBinding(nr, next)) = &e {
+                    let maybe_animation = animation_for_property(instance_ref, &binding.animation);
+                    for nr in &binding.two_way_bindings {
                         // Safety: The compiler must have ensured that the properties exist and are of the same type
-                        prop_rtti.link_two_ways(item, get_property_ptr(&nr, instance_ref));
-                        e = next.as_deref();
+                        prop_rtti.link_two_ways(item, get_property_ptr(nr, instance_ref));
                     }
-                    if let Some(e) = e {
-                        if is_const || e.is_constant() {
-                            prop_rtti.set(
-                                item,
-                                eval::eval_expression(
-                                    e,
-                                    &mut eval::EvalLocalContext::from_component_instance(
-                                        instance_ref,
+                    if !matches!(binding.expression, Expression::Invalid) {
+                        if is_const {
+                            prop_rtti
+                                .set(
+                                    item,
+                                    eval::eval_expression(
+                                        &binding.expression,
+                                        &mut eval::EvalLocalContext::from_component_instance(
+                                            instance_ref,
+                                        ),
                                     ),
-                                ),
-                                maybe_animation.as_animation(),
-                            );
+                                    maybe_animation.as_animation(),
+                                )
+                                .unwrap();
                         } else {
-                            let e = e.clone();
+                            let e = binding.expression.clone();
                             let component_type = component_type.clone();
                             let instance = component_box.instance.as_ptr();
                             let c = Pin::new_unchecked(vtable::VRef::from_raw(
@@ -1252,11 +1280,14 @@ pub(crate) fn get_property_ptr(nr: &NamedReference, instance: InstanceRef) -> *c
                 .items
                 .get(element.id.as_str())
                 .unwrap_or_else(|| panic!("Unknown element for {}.{}", element.id, nr.name()));
+            let prop_info = item_info
+                .rtti
+                .properties
+                .get(nr.name())
+                .unwrap_or_else(|| panic!("Property {} not in {}", nr.name(), element.id));
             core::mem::drop(element);
             let item = unsafe { item_info.item_from_component(enclosing_component.as_ptr()) };
-            unsafe {
-                item.as_ptr().add(item_info.rtti.properties.get(nr.name()).unwrap().offset()).cast()
-            }
+            unsafe { item.as_ptr().add(prop_info.offset()).cast() }
         }
         eval::ComponentInstance::GlobalComponent(glob) => glob.as_ref().get_property_ptr(nr.name()),
     }
@@ -1274,12 +1305,12 @@ impl ErasedComponentBox {
         )
     }
 
-    pub fn borrow<'a>(&'a self) -> ComponentRefPin<'a> {
+    pub fn borrow(&self) -> ComponentRefPin {
         // Safety: it is safe to access self.0 here because the 'id lifetime does not leak
         self.0.borrow()
     }
 
-    pub fn window(&self) -> ComponentWindow {
+    pub fn window(&self) -> &Window {
         self.0.window()
     }
 
@@ -1326,7 +1357,7 @@ extern "C" fn layout_info(component: ComponentRefPin, orientation: Orientation)
     let mut result = crate::eval_layout::get_layout_info(
         &instance_ref.component_type.original.root_element,
         instance_ref,
-        &eval::window_ref(instance_ref).unwrap(),
+        eval::window_ref(instance_ref).unwrap(),
         orientation,
     );
 
@@ -1367,7 +1398,7 @@ unsafe extern "C" fn parent_item(component: ComponentRefPin, index: usize, resul
             .original
             .parent_element
             .upgrade()
-            .and_then(|e| e.borrow().item_index.get().map(|x| *x));
+            .and_then(|e| e.borrow().item_index.get().cloned());
         if let (Some(parent_offset), Some(parent_index)) =
             (instance_ref.component_type.parent_component_offset, parent_item_index)
         {
@@ -1449,6 +1480,10 @@ impl<'a, 'id> InstanceRef<'a, 'id> {
         let extra_data = self.component_type.extra_data_offset.apply(self.as_ref());
         &extra_data.self_weak
     }
+
+    pub fn window(&self) -> &sixtyfps_corelib::window::api::Window {
+        self.component_type.window_offset.apply(self.as_ref()).as_ref().as_ref().unwrap()
+    }
 }
 
 /// Show the popup at the given location
@@ -1457,12 +1492,12 @@ pub fn show_popup(
     x: f32,
     y: f32,
     parent_comp: ComponentRefPin,
-    parent_window: ComponentWindow,
+    parent_window: &WindowRc,
 ) {
     generativity::make_guard!(guard);
     // FIXME: we should compile once and keep the cached compiled component
     let compiled = generate_component(&popup.component, guard);
-    let inst = instantiate(compiled, Some(parent_comp), Some(parent_window.clone()));
+    let inst = instantiate(compiled, Some(parent_comp), Some(parent_window));
     inst.run_setup_code();
     parent_window
         .show_popup(&vtable::VRc::into_dyn(inst), sixtyfps_corelib::graphics::Point::new(x, y));
