@@ -1,21 +1,20 @@
 /* LICENSE BEGIN
     This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2020 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2020 Simon Hausmann <simon.hausmann@sixtyfps.io>
+    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
+    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
 
     SPDX-License-Identifier: GPL-3.0-only
     This file is also available under commercial licensing terms.
     Please contact info@sixtyfps.io for more information.
 LICENSE END */
-use crate::api::{Struct, Value};
+use crate::api::{SetPropertyError, Struct, Value};
 use crate::dynamic_component::InstanceRef;
 use core::convert::TryInto;
-use core::iter::FromIterator;
 use core::pin::Pin;
 use corelib::graphics::{GradientStop, LinearGradientBrush, PathElement};
 use corelib::items::{ItemRef, PropertyAnimation};
 use corelib::rtti::AnimatedBindingKind;
-use corelib::window::ComponentWindow;
+use corelib::window::{WindowHandleAccess, WindowRc};
 use corelib::{Brush, Color, PathData, SharedString, SharedVector};
 use sixtyfps_compilerlib::expression_tree::{
     BindingExpression, BuiltinFunction, EasingCurve, Expression, Path as ExprPath,
@@ -29,7 +28,12 @@ use std::rc::Rc;
 
 pub trait ErasedPropertyInfo {
     fn get(&self, item: Pin<ItemRef>) -> Value;
-    fn set(&self, item: Pin<ItemRef>, value: Value, animation: Option<PropertyAnimation>);
+    fn set(
+        &self,
+        item: Pin<ItemRef>,
+        value: Value,
+        animation: Option<PropertyAnimation>,
+    ) -> Result<(), ()>;
     fn set_binding(
         &self,
         item: Pin<ItemRef>,
@@ -49,8 +53,13 @@ impl<Item: vtable::HasStaticVTable<corelib::items::ItemVTable>> ErasedPropertyIn
     fn get(&self, item: Pin<ItemRef>) -> Value {
         (*self).get(ItemRef::downcast_pin(item).unwrap()).unwrap()
     }
-    fn set(&self, item: Pin<ItemRef>, value: Value, animation: Option<PropertyAnimation>) {
-        (*self).set(ItemRef::downcast_pin(item).unwrap(), value, animation).unwrap()
+    fn set(
+        &self,
+        item: Pin<ItemRef>,
+        value: Value,
+        animation: Option<PropertyAnimation>,
+    ) -> Result<(), ()> {
+        (*self).set(ItemRef::downcast_pin(item).unwrap(), value, animation)
     }
     fn set_binding(
         &self,
@@ -128,14 +137,13 @@ impl<'a, 'id> EvalLocalContext<'a, 'id> {
 }
 
 /// Evaluate an expression and return a Value as the result of this expression
-pub fn eval_expression(e: &Expression, local_context: &mut EvalLocalContext) -> Value {
+pub fn eval_expression(expression: &Expression, local_context: &mut EvalLocalContext) -> Value {
     if let Some(r) = &local_context.return_value {
         return r.clone();
     }
-    match e {
+    match expression {
         Expression::Invalid => panic!("invalid expression while evaluating"),
         Expression::Uncompiled(_) => panic!("uncompiled expression while evaluating"),
-        Expression::TwoWayBinding(..) => panic!("invalid expression while evaluating"),
         Expression::StringLiteral(s) => Value::String(s.into()),
         Expression::NumberLiteral(n, unit) => Value::Number(unit.normalize(*n)),
         Expression::BoolLiteral(b) => Value::Bool(*b),
@@ -204,12 +212,12 @@ pub fn eval_expression(e: &Expression, local_context: &mut EvalLocalContext) ->
             }
             Expression::BuiltinFunctionReference(BuiltinFunction::Debug, _) => {
                 let to_print: SharedString = eval_expression(&arguments[0], local_context).try_into().unwrap();
-                eprintln!("{}", to_print);
+                corelib::debug_log!("{}", to_print);
                 Value::Void
             }
             Expression::BuiltinFunctionReference(BuiltinFunction::Mod, _) => {
-                let mut toint = |e| -> i32 { eval_expression(e, local_context).try_into().unwrap() };
-                Value::Number((toint(&arguments[0]) % toint(&arguments[1])) as _)
+                let mut to_int = |e| -> i32 { eval_expression(e, local_context).try_into().unwrap() };
+                Value::Number((to_int(&arguments[0]) % to_int(&arguments[1])) as _)
             }
             Expression::BuiltinFunctionReference(BuiltinFunction::Round, _) => {
                 let x: f64 = eval_expression(&arguments[0], local_context).try_into().unwrap();
@@ -275,7 +283,7 @@ pub fn eval_expression(e: &Expression, local_context: &mut EvalLocalContext) ->
 
                     let focus_item_comp = enclosing_component.self_weak().get().unwrap().upgrade().unwrap();
 
-                    window_ref(component).unwrap().set_focus_item(&corelib::items::ItemRc::new(vtable::VRc::into_dyn(focus_item_comp), item_info.item_index()));
+                    window_ref(component).unwrap().clone().set_focus_item(&corelib::items::ItemRc::new(vtable::VRc::into_dyn(focus_item_comp), item_info.item_index()));
                     Value::Void
                 } else {
                     panic!("internal error: argument to SetFocusItem must be an element")
@@ -393,7 +401,7 @@ pub fn eval_expression(e: &Expression, local_context: &mut EvalLocalContext) ->
                     let item_ref = unsafe { item_info.item_from_component(enclosing_component.as_ptr()) };
 
                     let window = window_ref(component).unwrap();
-                    item_ref.as_ref().layouting_info(crate::eval_layout::to_runtime(*orient), &window).into()
+                    item_ref.as_ref().layouting_info(crate::eval_layout::to_runtime(*orient), window).into()
                 } else {
                     panic!("internal error: incorrect arguments to ImplicitLayoutInfo {:?}", arguments);
                 }
@@ -469,7 +477,7 @@ pub fn eval_expression(e: &Expression, local_context: &mut EvalLocalContext) ->
                 sixtyfps_compilerlib::expression_tree::ImageReference::AbsolutePath(path) => {
                     corelib::graphics::Image::load_from_path(std::path::Path::new(path))
                 }
-                sixtyfps_compilerlib::expression_tree::ImageReference::EmbeddedData(_) => panic!("Resource embedding is not supported by the interpreter")
+                sixtyfps_compilerlib::expression_tree::ImageReference::EmbeddedData { .. } => panic!("Resource embedding is not supported by the interpreter")
             }.unwrap_or_else(|_| {
                 eprintln!("Could not load image {:?}",resource_ref );
                 Default::default()
@@ -522,7 +530,7 @@ pub fn eval_expression(e: &Expression, local_context: &mut EvalLocalContext) ->
             Value::EnumerationValue(value.enumeration.name.clone(), value.to_string())
         }
         Expression::ReturnStatement(x) => {
-            let val = x.as_ref().map_or(Value::Void, |x| eval_expression(&x, local_context));
+            let val = x.as_ref().map_or(Value::Void, |x| eval_expression(x, local_context));
             if local_context.return_value.is_none() {
                 local_context.return_value = Some(val);
             }
@@ -532,7 +540,7 @@ pub fn eval_expression(e: &Expression, local_context: &mut EvalLocalContext) ->
             let cache = load_property_helper(local_context.component_instance, &layout_cache_prop.element(), layout_cache_prop.name()).unwrap();
             if let Value::LayoutCache(cache) = cache {
                 if let Some(ri) = repeater_index {
-                    let offset : usize = eval_expression(&ri, local_context).try_into().unwrap();
+                    let offset : usize = eval_expression(ri, local_context).try_into().unwrap();
                     Value::Number(cache[(cache[*index] as usize) + offset * 2].into())
                 } else {
                     Value::Number(cache[*index].into())
@@ -594,12 +602,15 @@ fn eval_assignment(lhs: &Expression, op: char, rhs: Value, local_context: &mut E
                     let item =
                         unsafe { item_info.item_from_component(enclosing_component.as_ptr()) };
                     let p = &item_info.rtti.properties[nr.name()];
-                    p.set(item, eval(p.get(item)), None);
+                    p.set(item, eval(p.get(item)), None).unwrap();
                 }
                 ComponentInstance::GlobalComponent(global) => {
-                    let val =
-                        if op == '=' { rhs } else { eval(global.as_ref().get_property(nr.name())) };
-                    global.as_ref().set_property(nr.name(), val);
+                    let val = if op == '=' {
+                        rhs
+                    } else {
+                        eval(global.as_ref().get_property(nr.name()).unwrap())
+                    };
+                    global.as_ref().set_property(nr.name(), val).unwrap();
                 }
             }
         }
@@ -660,7 +671,7 @@ fn load_property_helper(
     name: &str,
 ) -> Result<Value, ()> {
     generativity::make_guard!(guard);
-    match enclosing_component_instance_for_element(&element, component_instance, guard) {
+    match enclosing_component_instance_for_element(element, component_instance, guard) {
         ComponentInstance::InstanceRef(enclosing_component) => {
             let element = element.borrow();
             if element.id == element.enclosing_component.upgrade().unwrap().root_element.borrow().id
@@ -669,6 +680,8 @@ fn load_property_helper(
                     return unsafe {
                         x.prop.get(Pin::new_unchecked(&*enclosing_component.as_ptr().add(x.offset)))
                     };
+                } else if enclosing_component.component_type.original.is_global() {
+                    return Err(());
                 }
             };
             let item_info = enclosing_component
@@ -680,7 +693,7 @@ fn load_property_helper(
             let item = unsafe { item_info.item_from_component(enclosing_component.as_ptr()) };
             Ok(item_info.rtti.properties.get(name).ok_or(())?.get(item))
         }
-        ComponentInstance::GlobalComponent(glob) => Ok(glob.as_ref().get_property(name)),
+        ComponentInstance::GlobalComponent(glob) => Ok(glob.as_ref().get_property(name).unwrap()),
     }
 }
 
@@ -689,13 +702,12 @@ pub fn store_property(
     element: &ElementRc,
     name: &str,
     value: Value,
-) -> Result<(), ()> {
+) -> Result<(), SetPropertyError> {
     generativity::make_guard!(guard);
-    let enclosing_component = enclosing_component_for_element(&element, component_instance, guard);
+    let enclosing_component = enclosing_component_for_element(element, component_instance, guard);
     let maybe_animation = crate::dynamic_component::animation_for_property(
         enclosing_component,
-        &element.borrow(),
-        name,
+        element.borrow().bindings.get(name).map_or(&None, |b| &b.animation),
     );
 
     let component = element.borrow().enclosing_component.upgrade().unwrap();
@@ -703,14 +715,19 @@ pub fn store_property(
         if let Some(x) = enclosing_component.component_type.custom_properties.get(name) {
             unsafe {
                 let p = Pin::new_unchecked(&*enclosing_component.as_ptr().add(x.offset));
-                return x.prop.set(p, value, maybe_animation.as_animation());
+                return x
+                    .prop
+                    .set(p, value, maybe_animation.as_animation())
+                    .map_err(|()| SetPropertyError::WrongType);
             }
+        } else if enclosing_component.component_type.original.is_global() {
+            return Err(SetPropertyError::NoSuchProperty);
         }
     };
     let item_info = &enclosing_component.component_type.items[element.borrow().id.as_str()];
     let item = unsafe { item_info.item_from_component(enclosing_component.as_ptr()) };
-    let p = &item_info.rtti.properties.get(name).ok_or(())?;
-    p.set(item, value, maybe_animation.as_animation());
+    let p = &item_info.rtti.properties.get(name).ok_or(SetPropertyError::NoSuchProperty)?;
+    p.set(item, value, maybe_animation.as_animation()).unwrap();
     Ok(())
 }
 
@@ -724,21 +741,44 @@ pub(crate) fn invoke_callback(
     match enclosing_component_instance_for_element(element, component_instance, guard) {
         ComponentInstance::InstanceRef(enclosing_component) => {
             let component_type = enclosing_component.component_type;
-            let item_info = &component_type.items[element.borrow().id.as_str()];
+            let element = element.borrow();
+            if element.id == element.enclosing_component.upgrade().unwrap().root_element.borrow().id
+            {
+                if let Some(callback_offset) = component_type.custom_callbacks.get(callback_name) {
+                    let callback = callback_offset.apply(&*enclosing_component.instance);
+                    let res = callback.call(args);
+                    return Some(if res != Value::Void {
+                        res
+                    } else if let Some(Type::Callback { return_type: Some(rt), .. }) =
+                        component_type
+                            .original
+                            .root_element
+                            .borrow()
+                            .property_declarations
+                            .get(callback_name)
+                            .map(|d| &d.property_type)
+                    {
+                        // If the callback was not se, the return value will be Value::Void, but we need
+                        // to make sure that the value is actually of the right type as returned by the
+                        // callback, otherwise we will get panics later
+                        default_value_for_type(&*rt)
+                    } else {
+                        res
+                    });
+                } else if enclosing_component.component_type.original.is_global() {
+                    return None;
+                }
+            };
+            let item_info = &component_type.items[element.id.as_str()];
             let item = unsafe { item_info.item_from_component(enclosing_component.as_ptr()) };
-
             if let Some(callback) = item_info.rtti.callbacks.get(callback_name) {
                 Some(callback.call(item, args))
-            } else if let Some(callback_offset) = component_type.custom_callbacks.get(callback_name)
-            {
-                let callback = callback_offset.apply(&*enclosing_component.instance);
-                Some(callback.call(args))
             } else {
                 None
             }
         }
         ComponentInstance::GlobalComponent(global) => {
-            Some(global.as_ref().invoke_callback(callback_name, args))
+            Some(global.as_ref().invoke_callback(callback_name, args).unwrap())
         }
     }
 }
@@ -769,8 +809,13 @@ fn root_component_instance<'a, 'old_id, 'new_id>(
     }
 }
 
-pub fn window_ref(component: InstanceRef) -> Option<ComponentWindow> {
-    component.component_type.window_offset.apply(&*component.instance.get_ref()).clone()
+pub fn window_ref<'a>(component: InstanceRef<'a, '_>) -> Option<&'a WindowRc> {
+    component
+        .component_type
+        .window_offset
+        .apply(component.instance.get_ref())
+        .as_ref()
+        .map(|window| window.window_handle())
 }
 
 /// Return the component instance which hold the given element.
@@ -821,7 +866,7 @@ pub(crate) fn enclosing_component_instance_for_element<'a, 'old_id, 'new_id>(
                 let root = root_component_instance(component, static_guard);
                 ComponentInstance::GlobalComponent(
                     &root.component_type.extra_data_offset.apply(&*root.instance.get_ref()).globals
-                        [enclosing.id.as_str()],
+                        [enclosing.root_element.borrow().id.as_str()],
                 )
             } else {
                 ComponentInstance::InstanceRef(enclosing_component_for_element(
@@ -843,7 +888,7 @@ pub fn new_struct_with_bindings<ElementType: 'static + Default + corelib::rtti::
     let mut element = ElementType::default();
     for (prop, info) in ElementType::fields::<Value>().into_iter() {
         if let Some(binding) = &bindings.get(prop) {
-            let value = eval_expression(&binding, local_context);
+            let value = eval_expression(binding, local_context);
             info.set_field(&mut element, value).unwrap();
         }
     }
@@ -895,15 +940,18 @@ fn convert_from_lyon_path<'a>(
 
     PathData::Events(
         SharedVector::from(events.as_slice()),
-        SharedVector::from_iter(coordinates.into_iter().cloned()),
+        coordinates.into_iter().cloned().collect::<SharedVector<_>>(),
     )
 }
 
 pub fn convert_path(path: &ExprPath, local_context: &mut EvalLocalContext) -> PathData {
     match path {
-        ExprPath::Elements(elements) => PathData::Elements(SharedVector::<PathElement>::from_iter(
-            elements.iter().map(|element| convert_path_element(element, local_context)),
-        )),
+        ExprPath::Elements(elements) => PathData::Elements(
+            elements
+                .iter()
+                .map(|element| convert_path_element(element, local_context))
+                .collect::<SharedVector<PathElement>>(),
+        ),
         ExprPath::Events(events) => convert_from_lyon_path(events.iter()),
     }
 }
@@ -949,9 +997,9 @@ pub fn default_value_for_type(ty: &Type) -> Value {
         Type::Image => Value::Image(Default::default()),
         Type::Bool => Value::Bool(false),
         Type::Callback { .. } => Value::Void,
-        Type::Struct { fields, .. } => Value::Struct(Struct::from_iter(
-            fields.iter().map(|(n, t)| (n.clone(), default_value_for_type(t))),
-        )),
+        Type::Struct { fields, .. } => Value::Struct(
+            fields.iter().map(|(n, t)| (n.clone(), default_value_for_type(t))).collect::<Struct>(),
+        ),
         Type::Array(_) => Value::Array(Default::default()),
         Type::Percent => Value::Number(0.),
         Type::Enumeration(e) => {
