@@ -1,7 +1,7 @@
 /* LICENSE BEGIN
     This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2020 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2020 Simon Hausmann <simon.hausmann@sixtyfps.io>
+    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
+    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
 
     SPDX-License-Identifier: GPL-3.0-only
     This file is also available under commercial licensing terms.
@@ -13,6 +13,7 @@ use crate::dynamic_component::ErasedComponentBox;
 use super::*;
 use sixtyfps_corelib::model::{Model, ModelNotify, ModelPeer};
 use sixtyfps_corelib::slice::Slice;
+use sixtyfps_corelib::window::{WindowHandleAccess, WindowRc};
 use std::ffi::c_void;
 use vtable::VRef;
 
@@ -117,6 +118,12 @@ pub unsafe extern "C" fn sixtyfps_interpreter_value_new_struct(
     std::ptr::write(val as *mut Value, Value::Struct(struc.as_struct().clone()))
 }
 
+/// Construct a new Value in the given memory location as image
+#[no_mangle]
+pub unsafe extern "C" fn sixtyfps_interpreter_value_new_image(img: &Image, val: *mut ValueOpaque) {
+    std::ptr::write(val as *mut Value, Value::Image(img.clone()))
+}
+
 /// Construct a new Value containing a model in the given memory location
 #[no_mangle]
 pub unsafe extern "C" fn sixtyfps_interpreter_value_new_model(
@@ -184,6 +191,14 @@ pub extern "C" fn sixtyfps_interpreter_value_to_struct(val: &ValueOpaque) -> Opt
     }
 }
 
+#[no_mangle]
+pub extern "C" fn sixtyfps_interpreter_value_to_image(val: &ValueOpaque) -> Option<&Image> {
+    match val.as_value() {
+        Value::Image(img) => Some(img),
+        _ => None,
+    }
+}
+
 #[repr(C)]
 #[cfg(target_pointer_width = "64")]
 pub struct StructOpaque([usize; 6]);
@@ -279,9 +294,9 @@ pub unsafe extern "C" fn sixtyfps_interpreter_struct_iterator_next<'a>(
 }
 
 #[no_mangle]
-pub extern "C" fn sixtyfps_interpreter_struct_make_iter<'a>(
-    stru: &'a StructOpaque,
-) -> StructIteratorOpaque<'a> {
+pub extern "C" fn sixtyfps_interpreter_struct_make_iter(
+    stru: &StructOpaque,
+) -> StructIteratorOpaque {
     let ret_it: StructIterator = stru.as_struct().0.iter();
     unsafe {
         let mut r = std::mem::MaybeUninit::<StructIteratorOpaque>::uninit();
@@ -301,7 +316,10 @@ pub unsafe extern "C" fn sixtyfps_interpreter_component_instance_get_property(
 ) -> bool {
     generativity::make_guard!(guard);
     let comp = inst.unerase(guard);
-    match comp.description().get_property(comp.borrow(), std::str::from_utf8(&name).unwrap()) {
+    match comp
+        .description()
+        .get_property(comp.borrow(), &normalize_identifier(std::str::from_utf8(&name).unwrap()))
+    {
         Ok(val) => {
             std::ptr::write(out as *mut Value, val);
             true
@@ -319,7 +337,11 @@ pub extern "C" fn sixtyfps_interpreter_component_instance_set_property(
     generativity::make_guard!(guard);
     let comp = inst.unerase(guard);
     comp.description()
-        .set_property(comp.borrow(), std::str::from_utf8(&name).unwrap(), val.as_value().clone())
+        .set_property(
+            comp.borrow(),
+            &normalize_identifier(std::str::from_utf8(&name).unwrap()),
+            val.as_value().clone(),
+        )
         .is_ok()
 }
 
@@ -338,7 +360,7 @@ pub unsafe extern "C" fn sixtyfps_interpreter_component_instance_invoke_callback
     let comp = inst.unerase(guard);
     match comp.description().invoke_callback(
         comp.borrow(),
-        std::str::from_utf8(&name).unwrap(),
+        &normalize_identifier(std::str::from_utf8(&name).unwrap()),
         args.as_slice(),
     ) {
         Ok(val) => {
@@ -349,6 +371,22 @@ pub unsafe extern "C" fn sixtyfps_interpreter_component_instance_invoke_callback
     }
 }
 
+/// Wrap the user_data provided by the native code and call the drop function on Drop.
+///
+/// Safety: user_data must be a pointer that can be destroyed by the drop_user_data function.
+struct CallbackUserData {
+    user_data: *mut c_void,
+    drop_user_data: Option<extern "C" fn(*mut c_void)>,
+}
+
+impl Drop for CallbackUserData {
+    fn drop(&mut self) {
+        if let Some(x) = self.drop_user_data {
+            x(self.user_data)
+        }
+    }
+}
+
 /// Set a handler for the callback.
 /// The `callback` function must initialize the `ret` (the `ret` passed to the callback is initialized and is assumed initialized after the function)
 #[no_mangle]
@@ -359,20 +397,7 @@ pub unsafe extern "C" fn sixtyfps_interpreter_component_instance_set_callback(
     user_data: *mut c_void,
     drop_user_data: Option<extern "C" fn(*mut c_void)>,
 ) -> bool {
-    struct UserData {
-        user_data: *mut c_void,
-        drop_user_data: Option<extern "C" fn(*mut c_void)>,
-    }
-
-    impl Drop for UserData {
-        fn drop(&mut self) {
-            if let Some(x) = self.drop_user_data {
-                x(self.user_data)
-            }
-        }
-    }
-    let ud = UserData { user_data, drop_user_data };
-
+    let ud = CallbackUserData { user_data, drop_user_data };
     let callback = move |args: &[Value]| -> Value {
         let args = std::mem::transmute::<&[Value], &[ValueOpaque]>(args);
         let mut ret = std::mem::MaybeUninit::<Value>::uninit();
@@ -385,12 +410,124 @@ pub unsafe extern "C" fn sixtyfps_interpreter_component_instance_set_callback(
     comp.description()
         .set_callback_handler(
             comp.borrow(),
-            std::str::from_utf8(&name).unwrap(),
+            &normalize_identifier(std::str::from_utf8(&name).unwrap()),
             Box::new(callback),
         )
         .is_ok()
 }
 
+/// Get a global property.
+/// The `out` parameter must be uninitialized. If this function returns true, the out will be initialized
+/// to the resulting value. If this function returns false, out is unchanged
+#[no_mangle]
+pub unsafe extern "C" fn sixtyfps_interpreter_component_instance_get_global_property(
+    inst: &ErasedComponentBox,
+    global: Slice<u8>,
+    property_name: Slice<u8>,
+    out: *mut ValueOpaque,
+) -> bool {
+    generativity::make_guard!(guard);
+    let comp = inst.unerase(guard);
+    match comp
+        .description()
+        .get_global(comp.borrow(), &normalize_identifier(std::str::from_utf8(&global).unwrap()))
+        .and_then(|g| {
+            g.as_ref()
+                .get_property(&normalize_identifier(std::str::from_utf8(&property_name).unwrap()))
+        }) {
+        Ok(val) => {
+            std::ptr::write(out as *mut Value, val);
+            true
+        }
+        Err(_) => false,
+    }
+}
+
+#[no_mangle]
+pub extern "C" fn sixtyfps_interpreter_component_instance_set_global_property(
+    inst: &ErasedComponentBox,
+    global: Slice<u8>,
+    property_name: Slice<u8>,
+    val: &ValueOpaque,
+) -> bool {
+    generativity::make_guard!(guard);
+    let comp = inst.unerase(guard);
+    comp.description()
+        .get_global(comp.borrow(), &normalize_identifier(std::str::from_utf8(&global).unwrap()))
+        .and_then(|g| {
+            g.as_ref()
+                .set_property(
+                    &normalize_identifier(std::str::from_utf8(&property_name).unwrap()),
+                    val.as_value().clone(),
+                )
+                .map_err(|_| ())
+        })
+        .is_ok()
+}
+
+/// The `callback` function must initialize the `ret` (the `ret` passed to the callback is initialized and is assumed initialized after the function)
+#[no_mangle]
+pub unsafe extern "C" fn sixtyfps_interpreter_component_instance_set_global_callback(
+    inst: &ErasedComponentBox,
+    global: Slice<u8>,
+    name: Slice<u8>,
+    callback: extern "C" fn(user_data: *mut c_void, arg: Slice<ValueOpaque>, ret: *mut ValueOpaque),
+    user_data: *mut c_void,
+    drop_user_data: Option<extern "C" fn(*mut c_void)>,
+) -> bool {
+    let ud = CallbackUserData { user_data, drop_user_data };
+
+    let callback = move |args: &[Value]| -> Value {
+        let args = std::mem::transmute::<&[Value], &[ValueOpaque]>(args);
+        let mut ret = std::mem::MaybeUninit::<Value>::uninit();
+        callback(ud.user_data, Slice::from_slice(args), ret.as_mut_ptr() as *mut ValueOpaque);
+        ret.assume_init()
+    };
+
+    generativity::make_guard!(guard);
+    let comp = inst.unerase(guard);
+    comp.description()
+        .get_global(comp.borrow(), &normalize_identifier(std::str::from_utf8(&global).unwrap()))
+        .and_then(|g| {
+            g.as_ref().set_callback_handler(
+                &normalize_identifier(std::str::from_utf8(&name).unwrap()),
+                Box::new(callback),
+            )
+        })
+        .is_ok()
+}
+
+/// Invoke a global callback.
+/// The `out` parameter must be uninitialized. If this function returns true, the out will be initialized
+/// to the resulting value. If this function returns false, out is unchanged
+#[no_mangle]
+pub unsafe extern "C" fn sixtyfps_interpreter_component_instance_invoke_global_callback(
+    inst: &ErasedComponentBox,
+    global: Slice<u8>,
+    callback_name: Slice<u8>,
+    args: Slice<ValueOpaque>,
+    out: *mut ValueOpaque,
+) -> bool {
+    let args = std::mem::transmute::<Slice<ValueOpaque>, Slice<Value>>(args);
+    generativity::make_guard!(guard);
+    let comp = inst.unerase(guard);
+    match comp
+        .description()
+        .get_global(comp.borrow(), &normalize_identifier(std::str::from_utf8(&global).unwrap()))
+        .and_then(|g| {
+            g.as_ref().invoke_callback(
+                &normalize_identifier(std::str::from_utf8(&callback_name).unwrap()),
+                args.as_slice(),
+            )
+        }) {
+        Ok(val) => {
+            std::ptr::write(out as *mut Value, val);
+            true
+        }
+        Err(_) => false,
+    }
+}
+
 /// Show or hide
 #[no_mangle]
 pub extern "C" fn sixtyfps_interpreter_component_instance_show(
@@ -400,27 +537,26 @@ pub extern "C" fn sixtyfps_interpreter_component_instance_show(
     generativity::make_guard!(guard);
     let comp = inst.unerase(guard);
     if is_visible {
-        comp.window().show();
+        comp.borrow_instance().window().show();
     } else {
-        comp.window().hide();
+        comp.borrow_instance().window().hide();
     }
 }
 
 /// Return a window for the component
 ///
 /// The out pointer must be uninitialized and must be destroyed with
-/// sixtyfps_component_window_drop after usage
+/// sixtyfps_windowrc_drop after usage
 #[no_mangle]
 pub unsafe extern "C" fn sixtyfps_interpreter_component_instance_window(
     inst: &ErasedComponentBox,
-    out: *mut sixtyfps_corelib::window::ffi::ComponentWindowOpaque,
+    out: *mut *const sixtyfps_corelib::window::ffi::WindowRcOpaque,
 ) {
-    use sixtyfps_corelib::window::ComponentWindow;
     assert_eq!(
-        core::mem::size_of::<ComponentWindow>(),
-        core::mem::size_of::<sixtyfps_corelib::window::ffi::ComponentWindowOpaque>()
+        core::mem::size_of::<WindowRc>(),
+        core::mem::size_of::<sixtyfps_corelib::window::ffi::WindowRcOpaque>()
     );
-    core::ptr::write(out as *mut ComponentWindow, inst.window())
+    core::ptr::write(out as *mut *const WindowRc, inst.window().window_handle() as *const _)
 }
 
 /// Instantiate an instance from a definition.
@@ -568,7 +704,7 @@ impl ComponentCompilerOpaque {
 pub unsafe extern "C" fn sixtyfps_interpreter_component_compiler_new(
     compiler: *mut ComponentCompilerOpaque,
 ) {
-    std::ptr::write(compiler as *mut ComponentCompiler, ComponentCompiler::new())
+    std::ptr::write(compiler as *mut ComponentCompiler, ComponentCompiler::default())
 }
 
 #[no_mangle]
@@ -615,7 +751,7 @@ pub unsafe extern "C" fn sixtyfps_interpreter_component_compiler_get_include_pat
             .as_component_compiler()
             .include_paths()
             .iter()
-            .map(|path| path.to_str().map_or_else(|| Default::default(), |str| str.into())),
+            .map(|path| path.to_str().map_or_else(Default::default, |str| str.into())),
     );
 }
 
@@ -631,7 +767,7 @@ pub unsafe extern "C" fn sixtyfps_interpreter_component_compiler_get_diagnostics
             source_file: diagnostic
                 .source_file()
                 .and_then(|path| path.to_str())
-                .map_or_else(|| Default::default(), |str| str.into()),
+                .map_or_else(Default::default, |str| str.into()),
             line,
             column,
             level: match diagnostic.level() {
