@@ -1,7 +1,7 @@
 /* LICENSE BEGIN
     This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2020 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2020 Simon Hausmann <simon.hausmann@sixtyfps.io>
+    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
+    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
 
     SPDX-License-Identifier: GPL-3.0-only
     This file is also available under commercial licensing terms.
@@ -13,6 +13,7 @@ use std::rc::Rc;
 
 use crate::api::Value;
 use crate::dynamic_component::{ErasedComponentBox, ErasedComponentDescription};
+use crate::SetPropertyError;
 use sixtyfps_compilerlib::namedreference::NamedReference;
 use sixtyfps_compilerlib::{langtype::Type, object_tree::Component};
 use sixtyfps_corelib::component::ComponentVTable;
@@ -23,13 +24,36 @@ pub enum CompiledGlobal {
     Component(ErasedComponentDescription),
 }
 
-pub trait GlobalComponent {
-    fn invoke_callback(self: Pin<&Self>, _callback_name: &str, _args: &[Value]) -> Value {
-        todo!("call callback")
+impl CompiledGlobal {
+    pub fn names(&self) -> Vec<String> {
+        match self {
+            CompiledGlobal::Builtin(name, _) => vec![name.clone()],
+            CompiledGlobal::Component(component) => {
+                generativity::make_guard!(guard);
+                let component = component.unerase(guard);
+                let mut names = component.original.global_aliases();
+                names.push(component.original.root_element.borrow().id.clone());
+                names
+            }
+        }
     }
+}
 
-    fn set_property(self: Pin<&Self>, prop_name: &str, value: Value);
-    fn get_property(self: Pin<&Self>, prop_name: &str) -> Value;
+pub trait GlobalComponent {
+    fn invoke_callback(self: Pin<&Self>, callback_name: &str, args: &[Value]) -> Result<Value, ()>;
+
+    fn set_callback_handler(
+        self: Pin<&Self>,
+        callback_name: &str,
+        handler: Box<dyn Fn(&[Value]) -> Value>,
+    ) -> Result<(), ()>;
+
+    fn set_property(
+        self: Pin<&Self>,
+        prop_name: &str,
+        value: Value,
+    ) -> Result<(), SetPropertyError>;
+    fn get_property(self: Pin<&Self>, prop_name: &str) -> Result<Value, ()>;
 
     fn get_property_ptr(self: Pin<&Self>, prop_name: &str) -> *const ();
 }
@@ -66,7 +90,12 @@ pub fn instantiate(description: &CompiledGlobal) -> (String, Pin<Rc<dyn GlobalCo
                 None,
                 None,
             )));
-            (component.id.clone(), g)
+            let id = if component.is_global() {
+                component.root_element.borrow().id.clone()
+            } else {
+                component.id.clone()
+            };
+            (id, g)
         }
     }
 }
@@ -76,16 +105,20 @@ pub fn instantiate(description: &CompiledGlobal) -> (String, Pin<Rc<dyn GlobalCo
 pub struct GlobalComponentInstance(vtable::VRc<ComponentVTable, ErasedComponentBox>);
 
 impl GlobalComponent for GlobalComponentInstance {
-    fn set_property(self: Pin<&Self>, prop_name: &str, value: Value) {
+    fn set_property(
+        self: Pin<&Self>,
+        prop_name: &str,
+        value: Value,
+    ) -> Result<(), SetPropertyError> {
         generativity::make_guard!(guard);
         let comp = self.0.unerase(guard);
-        comp.description().set_property(comp.borrow(), prop_name, value).unwrap()
+        comp.description().set_property(comp.borrow(), prop_name, value)
     }
 
-    fn get_property(self: Pin<&Self>, prop_name: &str) -> Value {
+    fn get_property(self: Pin<&Self>, prop_name: &str) -> Result<Value, ()> {
         generativity::make_guard!(guard);
         let comp = self.0.unerase(guard);
-        comp.description().get_property(comp.borrow(), prop_name).unwrap()
+        comp.description().get_property(comp.borrow(), prop_name)
     }
 
     fn get_property_ptr(self: Pin<&Self>, prop_name: &str) -> *const () {
@@ -96,17 +129,41 @@ impl GlobalComponent for GlobalComponentInstance {
             comp.borrow_instance(),
         )
     }
+
+    fn invoke_callback(self: Pin<&Self>, callback_name: &str, args: &[Value]) -> Result<Value, ()> {
+        generativity::make_guard!(guard);
+        let comp = self.0.unerase(guard);
+        comp.description().invoke_callback(comp.borrow(), callback_name, args)
+    }
+
+    fn set_callback_handler(
+        self: Pin<&Self>,
+        callback_name: &str,
+        handler: Box<dyn Fn(&[Value]) -> Value>,
+    ) -> Result<(), ()> {
+        generativity::make_guard!(guard);
+        let comp = self.0.unerase(guard);
+        comp.description().set_callback_handler(comp.borrow(), callback_name, handler)
+    }
 }
 
 impl<T: rtti::BuiltinItem + 'static> GlobalComponent for T {
-    fn set_property(self: Pin<&Self>, prop_name: &str, value: Value) {
-        let prop = Self::properties().into_iter().find(|(k, _)| *k == prop_name).unwrap().1;
-        prop.set(self, value, None).unwrap()
+    fn set_property(
+        self: Pin<&Self>,
+        prop_name: &str,
+        value: Value,
+    ) -> Result<(), SetPropertyError> {
+        let prop = Self::properties()
+            .into_iter()
+            .find(|(k, _)| *k == prop_name)
+            .ok_or(SetPropertyError::NoSuchProperty)?
+            .1;
+        prop.set(self, value, None).map_err(|()| SetPropertyError::WrongType)
     }
 
-    fn get_property(self: Pin<&Self>, prop_name: &str) -> Value {
-        let prop = Self::properties().into_iter().find(|(k, _)| *k == prop_name).unwrap().1;
-        prop.get(self).unwrap()
+    fn get_property(self: Pin<&Self>, prop_name: &str) -> Result<Value, ()> {
+        let prop = Self::properties().into_iter().find(|(k, _)| *k == prop_name).ok_or(())?.1;
+        prop.get(self)
     }
 
     fn get_property_ptr(self: Pin<&Self>, prop_name: &str) -> *const () {
@@ -114,6 +171,20 @@ impl<T: rtti::BuiltinItem + 'static> GlobalComponent for T {
             Self::properties().into_iter().find(|(k, _)| *k == prop_name).unwrap().1;
         unsafe { (self.get_ref() as *const Self as *const u8).add(prop.offset()) as *const () }
     }
+
+    fn invoke_callback(self: Pin<&Self>, callback_name: &str, args: &[Value]) -> Result<Value, ()> {
+        let cb = Self::callbacks().into_iter().find(|(k, _)| *k == callback_name).ok_or(())?.1;
+        cb.call(self, args)
+    }
+
+    fn set_callback_handler(
+        self: Pin<&Self>,
+        callback_name: &str,
+        handler: Box<dyn Fn(&[Value]) -> Value>,
+    ) -> Result<(), ()> {
+        let cb = Self::callbacks().into_iter().find(|(k, _)| *k == callback_name).ok_or(())?.1;
+        cb.set_handler(self, handler)
+    }
 }
 
 pub(crate) fn generate(component: &Rc<Component>) -> CompiledGlobal {
