@@ -1,7 +1,7 @@
 /* LICENSE BEGIN
     This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2020 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2020 Simon Hausmann <simon.hausmann@sixtyfps.io>
+    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
+    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
 
     SPDX-License-Identifier: GPL-3.0-only
     This file is also available under commercial licensing terms.
@@ -32,7 +32,7 @@ This example loads a `.60` dynamically from a path and show errors if any:
 ```rust
 use sixtyfps_interpreter::{ComponentDefinition, ComponentCompiler};
 
-let mut compiler = ComponentCompiler::new();
+let mut compiler = ComponentCompiler::default();
 let definition =
     spin_on::spin_on(compiler.build_from_path("hello.60"));
 # #[cfg(feature="print_diagnostics")]
@@ -57,7 +57,7 @@ let code = r#"
     }
 "#;
 
-let mut compiler = ComponentCompiler::new();
+let mut compiler = ComponentCompiler::default();
 let definition =
     spin_on::spin_on(compiler.build_from_source(code.into(), Default::default()));
 assert!(compiler.diagnostics().is_empty());
@@ -69,7 +69,7 @@ instance.run();
 
 ## Features
 
-**display-diagnostics**: enable the `[print_diagnostics]` function to show diagnostic in the console output
+**display-diagnostics**: enable the [`print_diagnostics`] function to show diagnostic in the console output
 */
 #![warn(missing_docs)]
 #![doc(html_logo_url = "https://sixtyfps.io/resources/logo.drawio.svg")]
